/* Copyright (C) 2013-2016 Vincent Forest (vaplv@free.fr)
 *
 * The RSys library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The RSys library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the RSys library. If not, see <http://www.gnu.org/licenses/>. */

#define _POSIX_C_SOURCE 200112L /* snprintf support */
#include "image.h"
#include <stdio.h>
#include <string.h>

#ifdef COMPILER_CL
  #include "io_c99.h"
#endif

res_T
image_ppm_write
  (const char* path,
   const int width,
   const int height,
   const int Bpp,
   const unsigned char* buffer)
{
  FILE* fp = NULL;
  res_T res = RES_OK;

  fp = fopen(path, "w");
  if(NULL == fp) {
    res = RES_IO_ERR;
    goto error;
  }

  res = image_ppm_write_stream(fp, width, height, Bpp, buffer);
  if(res != RES_OK) goto error;

exit:
  if(fp) fclose(fp);
  return res;
error:
  goto exit;
}

res_T
image_ppm_write_stream
  (FILE* fp,
   const int width,
   const int height,
   const int Bpp,
   const unsigned char* buffer)
{
  char buf[BUFSIZ];
  res_T res = RES_OK;

  if(width && height && Bpp && !buffer) {
    res = RES_BAD_ARG;
    goto error;
  }
  if(!fp) {
    res = RES_BAD_ARG;
    goto error;
  }

  #define FWRITE(Fp, String)                                                   \
    {                                                                          \
      const size_t i = fwrite(String, sizeof(char), strlen(String), Fp);       \
      if(i != strlen(String) * sizeof(char)) { res = RES_MEM_ERR; goto error; }\
    } (void)0
  #define SNPRINTF(Buf, Sz, Str, Arg0, Arg1, Arg2)                             \
    {                                                                          \
      const int i = snprintf(Buf, Sz, Str, Arg0, Arg1, Arg2);                  \
      if( i >= BUFSIZ ) { res = RES_MEM_ERR; goto error; }                     \
    } (void)0

  SNPRINTF(buf, BUFSIZ, "P3\n\n%i %i\n%i\n", width, height, 255);
  FWRITE(fp, buf);

  if(Bpp) {
    const long pitch = width * Bpp;
    int x, y;
    for(y = 0; y < height; ++y) {
      const unsigned char* row = buffer + y * pitch;
      for(x = 0; x < width; ++x) {
        const unsigned char* pixel = row + x * Bpp;
        SNPRINTF(buf, BUFSIZ, "%u %u %u\n",
          pixel[0],
          Bpp > 1 ? pixel[1] : pixel[0],
          Bpp > 2 ? pixel[2] : pixel[0]);
        FWRITE(fp, buf);
      }
      FWRITE(fp, "\n");
    }
  }
  #undef SNPRINTF
  #undef FWRITE

exit:
  return res;
error:
  goto exit;
}

