/* Copyright (C) 2013-2016 Vincent Forest (vaplv@free.fr)
 *
 * The RSys library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The RSys library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the RSys library. If not, see <http://www.gnu.org/licenses/>. */

#include "float44.h"
#include "rsys.h"
#include <float.h>

#define CHECK_F4(A, B)                                                         \
  {                                                                            \
    const float* a__ = (A);                                                    \
    const float* b__ = (B);                                                    \
    int i__;                                                                   \
    FOR_EACH(i__, 0, 4)                                                        \
      CHECK(a__[i__], b__[i__]);                                               \
  } (void)0
#define CHECK_F44(A, B)                                                        \
  {                                                                            \
    const float* a__ = (A);                                                    \
    const float* b__ = (B);                                                    \
    int i__;                                                                   \
    FOR_EACH(i__, 0, 16)                                                       \
      CHECK(a__[i__], b__[i__]);                                               \
  } (void)0

int
main(int argc, char** argv)
{
  float a[16], b[16], dst[16], c[16];
  int i;
  (void)argc, (void)argv;

  f44_set(a, f44_splat(c, -1.f));
  FOR_EACH(i, 0, 16) {
    CHECK(a[i], -1.f);
    CHECK(c[i], -1.f);
  }
  f44
    (a,
     0.f, 1.f, 2.f, 3.f,
     4.f, 5.f, 6.f, 7.f,
     8.f, 9.f, 10.f, 11.f,
     12.f, 13.f, 14.f, 15.f);
  FOR_EACH(i, 0, 16) {
    CHECK(a[i], (float)i);
  }
  CHECK_F44
    (f44_set_identity(a),f44(c,
     1.f, 0.f, 0.f, 0.f,
     0.f, 1.f, 0.f, 0.f,
     0.f, 0.f, 1.f, 0.f,
     0.f, 0.f, 0.f, 1.f));
  CHECK_F44
    (a, f44(c,
     1.f, 0.f, 0.f, 0.f,
     0.f, 1.f, 0.f, 0.f,
     0.f, 0.f, 1.f, 0.f,
     0.f, 0.f, 0.f, 1.f));

  f44_splat(a, -1.f);
  CHECK_F44
    (f44_set_row(a, f4(c, 0.f, 1.f, 2.f, 3.f), 0), f44(c,
     0.f, -1.f, -1.f, -1.f,
     1.f, -1.f, -1.f, -1.f,
     2.f, -1.f, -1.f, -1.f,
     3.f, -1.f, -1.f, -1.f));
  CHECK_F44
    (a, f44(c,
     0.f, -1.f, -1.f, -1.f,
     1.f, -1.f, -1.f, -1.f,
     2.f, -1.f, -1.f, -1.f,
     3.f, -1.f, -1.f, -1.f));
  CHECK_F44
    (f44_set_row(a, f4(c, 4.f, 5.f, 6.f, 7.f), 1), f44(c,
     0.f, 4.f, -1.f, -1.f,
     1.f, 5.f, -1.f, -1.f,
     2.f, 6.f, -1.f, -1.f,
     3.f, 7.f, -1.f, -1.f));
  CHECK_F44
    (a, f44(c,
     0.f, 4.f, -1.f, -1.f,
     1.f, 5.f, -1.f, -1.f,
     2.f, 6.f, -1.f, -1.f,
     3.f, 7.f, -1.f, -1.f));
  CHECK_F44
    (f44_set_row(a, f4(c, 8.f, 9.f, 10.f, 11.f), 2),f44(c,
     0.f, 4.f, 8.f, -1.f,
     1.f, 5.f, 9.f, -1.f,
     2.f, 6.f, 10.f, -1.f,
     3.f, 7.f, 11.f, -1.f));
  CHECK_F44
    (a, f44(c,
     0.f, 4.f, 8.f, -1.f,
     1.f, 5.f, 9.f, -1.f,
     2.f, 6.f, 10.f, -1.f,
     3.f, 7.f, 11.f, -1.f));
  CHECK_F44
    (f44_set_row(a, f4(c, 12.f, 13.f, 14.f, 15.f), 3), f44(c,
     0.f, 4.f, 8.f, 12.f,
     1.f, 5.f, 9.f, 13.f,
     2.f, 6.f, 10.f, 14.f,
     3.f, 7.f, 11.f, 15.f));
  CHECK_F44
    (a, f44(c,
     0.f, 4.f, 8.f, 12.f,
     1.f, 5.f, 9.f, 13.f,
     2.f, 6.f, 10.f, 14.f,
     3.f, 7.f, 11.f, 15.f));

  CHECK_F44
    (f44_transpose(a, a), f44(c,
     0.f, 1.f, 2.f, 3.f,
     4.f, 5.f, 6.f, 7.f,
     8.f, 9.f, 10.f, 11.f,
     12.f, 13.f, 14.f, 15.f));
  CHECK_F44
    (a, f44(c,
     0.f, 1.f, 2.f, 3.f,
     4.f, 5.f, 6.f, 7.f,
     8.f, 9.f, 10.f, 11.f,
     12.f, 13.f, 14.f, 15.f));
  CHECK_F44
    (f44_transpose(b, a),f44(c,
     0.f, 4.f, 8.f, 12.f,
     1.f, 5.f, 9.f, 13.f,
     2.f, 6.f, 10.f, 14.f,
     3.f, 7.f, 11.f, 15.f));
  CHECK_F44
    (b, f44(c,
     0.f, 4.f, 8.f, 12.f,
     1.f, 5.f, 9.f, 13.f,
     2.f, 6.f, 10.f, 14.f,
     3.f, 7.f, 11.f, 15.f));

  f44_splat(a, -1.f);
  CHECK_F44
    (f44_set_col(a, f4(c, 0.f, 1.f, 2.f, 3.f), 0), f44(c,
     0.f, 1.f, 2.f, 3.f,
     -1.f, -1.f, -1.f, -1.f,
     -1.f, -1.f, -1.f, -1.f,
     -1.f, -1.f, -1.f, -1.f));
  CHECK_F44
    (a, f44(c,
     0.f, 1.f, 2.f, 3.f,
     -1.f, -1.f, -1.f, -1.f,
     -1.f, -1.f, -1.f, -1.f,
     -1.f, -1.f, -1.f, -1.f));
  CHECK_F44
    (f44_set_col(a, f4(c, 4.f, 5.f, 6.f, 7.f), 1), f44(c,
     0.f, 1.f, 2.f, 3.f,
     4.f, 5.f, 6.f, 7.f,
     -1.f, -1.f, -1.f, -1.f,
     -1.f, -1.f, -1.f, -1.f));
  CHECK_F44
    (a, f44(c,
     0.f, 1.f, 2.f, 3.f,
     4.f, 5.f, 6.f, 7.f,
     -1.f, -1.f, -1.f, -1.f,
     -1.f, -1.f, -1.f, -1.f));
  CHECK_F44
    (f44_set_col(a, f4(c, 8.f, 9.f, 10.f, 11.f), 2), f44(c,
     0.f, 1.f, 2.f, 3.f,
     4.f, 5.f, 6.f, 7.f,
     8.f, 9.f, 10.f, 11.f,
     -1.f, -1.f, -1.f, -1.f));
  CHECK_F44
    (a, f44(c,
     0.f, 1.f, 2.f, 3.f,
     4.f, 5.f, 6.f, 7.f,
     8.f, 9.f, 10.f, 11.f,
     -1.f, -1.f, -1.f, -1.f));
  CHECK_F44
    (f44_set_col(a, f4(c, 12.f, 13.f, 14.f, 15.f), 3),f44(c,
     0.f, 1.f, 2.f, 3.f,
     4.f, 5.f, 6.f, 7.f,
     8.f, 9.f, 10.f, 11.f,
     12.f, 13.f, 14.f, 15.f));
  CHECK_F44
    (a, f44(c,
     0.f, 1.f, 2.f, 3.f,
     4.f, 5.f, 6.f, 7.f,
     8.f, 9.f, 10.f, 11.f,
     12.f, 13.f, 14.f, 15.f));

  CHECK_F4(f44_row(b + 1, a, 0), f4(c, 0.f, 4.f, 8.f, 12.f));
  CHECK_F4(b + 1, f4(c, 0.f, 4.f, 8.f, 12.f));
  CHECK_F4(f44_row(b + 2, a, 1), f4(c, 1.f, 5.f, 9.f, 13.f));
  CHECK_F4(b + 2, f4(c, 1.f, 5.f, 9.f, 13.f));
  CHECK_F4(f44_row(b + 8, a, 2), f4(c, 2.f, 6.f, 10.f, 14.f));
  CHECK_F4(b + 8, f4(c, 2.f, 6.f, 10.f, 14.f));
  CHECK_F4(f44_row(b + 5, a, 2), f4(c, 2.f, 6.f, 10.f, 14.f));
  CHECK_F4(b + 5, f4(c, 2.f, 6.f, 10.f, 14.f));

  CHECK_F4(f44_col(b + 2, a, 0), f4(c, 0.f, 1.f, 2.f, 3.f));
  CHECK_F4(b + 2, f4(c, 0.f, 1.f, 2.f, 3.f));
  CHECK_F4(f44_col(b + 1, a, 1), f4(c, 4.f, 5.f, 6.f, 7.f));
  CHECK_F4(b + 1, f4(c, 4.f, 5.f, 6.f, 7.f));
  CHECK_F4(f44_col(b + 5, a, 2), f4(c, 8.f, 9.f, 10.f, 11.f));
  CHECK_F4(b + 5, f4(c, 8.f, 9.f, 10.f, 11.f));
  CHECK_F4(f44_col(b + 5, a, 3), f4(c, 12.f, 13.f, 14.f, 15.f));
  CHECK_F4(b + 5, f4(c, 12.f, 13.f, 14.f, 15.f));
  CHECK_F4(f44_col_ptr(a, 0), f4(c, 0.f, 1.f, 2.f, 3.f));
  CHECK_F4(f44_col_ptr(a, 1), f4(c, 4.f, 5.f, 6.f, 7.f));
  CHECK_F4(f44_col_ptr(a, 2), f4(c, 8.f, 9.f, 10.f, 11.f));
  CHECK_F4(f44_col_ptr(a, 3), f4(c, 12.f, 13.f, 14.f, 15.f));
  CHECK_F4(f44_col_cptr(a, 0), f4(c, 0.f, 1.f, 2.f, 3.f));
  CHECK_F4(f44_col_cptr(a, 1), f4(c, 4.f, 5.f, 6.f, 7.f));
  CHECK_F4(f44_col_cptr(a, 2), f4(c, 8.f, 9.f, 10.f, 11.f));
  CHECK_F4(f44_col_cptr(a, 3), f4(c, 12.f, 13.f, 14.f, 15.f));

  CHECK_F4
    (f44_mulf4(dst, a, f4(c, 1.f, 2.f, 3.f, 1.f)),
     f4(c, 44.f, 51.f, 58.f, 65.f));
  CHECK_F4(dst, f4(c, 44.f, 51.f, 58.f, 65.f));
  CHECK_F4
    (f4_mulf44(dst, f4(c, 1.f, 2.f, 3.f, 1.f), a),
     f4(c, 11.f, 39.f, 67.f, 95.f));
  CHECK_F4(dst, f4(c, 11.f, 39.f, 67.f, 95.f));
  CHECK_F44
    (f44_mul(dst, a, -1.f), f44(c,
     0.f, -1.f, -2.f, -3.f,
     -4.f, -5.f, -6.f, -7.f,
     -8.f, -9.f, -10.f, -11.f,
     -12.f, -13.f, -14.f, -15.f));
  CHECK_F44
    (dst, f44(c,
     0.f, -1.f, -2.f, -3.f,
     -4.f, -5.f, -6.f, -7.f,
     -8.f, -9.f, -10.f, -11.f,
     -12.f, -13.f, -14.f, -15.f));

  f44
    (a,
     1.f, 2.f, 3.f, 4.f,
     4.f, 5.f, 6.f, 7.f,
     7.f, 8.f, 9.f, 10.f,
     10.f, 11.f, 12.f, 13.f);
  f44
    (b,
     2.f, 9.f, 8.f, 1.f,
     1.f, -2.f, 2.f, 1.f,
     1.f, -8.f, -4.f, 2.f,
     1.f, 3.f, 4.f, 2.f);
  CHECK_F44
    (f44_mulf44(dst, a, b), f44(c,
     104.f, 124.f, 144.f, 164.f,
     17.f, 19.f, 21.f, 23.f,
     -39.f, -48.f, -57.f, -66.f,
     61.f, 71.f, 81.f, 91.f));
  CHECK_F44
    (dst, f44(c,
     104.f, 124.f, 144.f, 164.f,
     17.f, 19.f, 21.f, 23.f,
     -39.f, -48.f, -57.f, -66.f,
     61.f, 71.f, 81.f, 91.f));
  CHECK_F44
    (f44_mulf44(a, a, b), f44(c,
     104.f, 124.f, 144.f, 164.f,
     17.f, 19.f, 21.f, 23.f,
     -39.f, -48.f, -57.f, -66.f,
     61.f, 71.f, 81.f, 91.f));
  CHECK_F44
    (a, f44(c,
     104.f, 124.f, 144.f, 164.f,
     17.f, 19.f, 21.f, 23.f,
     -39.f, -48.f, -57.f, -66.f,
     61.f, 71.f, 81.f, 91.f));
  f44
    (a,
     1.f, 2.f, 3.f, 4.f,
     4.f, 5.f, 6.f, 7.f,
     7.f, 8.f, 9.f, 10.f,
     10.f, 11.f, 12.f, 13.f);
  CHECK_F44
    (f44_mulf44(b, a, b), f44(c,
     104.f, 124.f, 144.f, 164.f,
     17.f, 19.f, 21.f, 23.f,
     -39.f, -48.f, -57.f, -66.f,
     61.f, 71.f, 81.f, 91.f));
  CHECK_F44
    (b, f44(c,
     104.f, 124.f, 144.f, 164.f,
     17.f, 19.f, 21.f, 23.f,
     -39.f, -48.f, -57.f, -66.f,
     61.f, 71.f, 81.f, 91.f));

  f44
    (a,
     0.f, 1.f, 2.f, 3.f,
     4.f, 5.f, 6.f, 7.f,
     8.f, 9.f, 10.f, 11.f,
     12.f, 13.f, 14.f, 15.f);
  f44
    (b,
     0.f, 2.f, 1.f, 3.f,
     1.f, -2.f, -1.f, -3.f,
     1.f, 0.f, 0.f, 2.f,
     3.f, 2.f, 1.f, 0.f);
  CHECK_F44
    (f44_add(dst, a, b), f44(c,
     0.f, 3.f, 3.f, 6.f,
     5.f, 3.f, 5.f, 4.f,
     9.f, 9.f, 10.f, 13.f,
     15.f, 15.f, 15.f, 15.f));
  CHECK_F44
    (dst, f44(c,
     0.f, 3.f, 3.f, 6.f,
     5.f, 3.f, 5.f, 4.f,
     9.f, 9.f, 10.f, 13.f,
     15.f, 15.f, 15.f, 15.f));
  CHECK_F44
    (f44_sub(dst, a, b), f44(c,
     0.f, -1.f, 1.f, 0.f,
     3.f, 7.f, 7.f, 10.f,
     7.f, 9.f, 10.f, 9.f,
     9.f, 11.f, 13.f, 15.f));
  CHECK_F44
    (dst, f44(c,
     0.f, -1.f, 1.f, 0.f,
     3.f, 7.f, 7.f, 10.f,
     7.f, 9.f, 10.f, 9.f,
     9.f, 11.f, 13.f, 15.f));
  CHECK_F44
    (f44_minus(a, b), f44(c,
     0.f, -2.f, -1.f, -3.f,
     -1.f, 2.f, 1.f, 3.f,
     -1.f, 0.f, 0.f, -2.f,
     -3.f, -2.f, -1.f, 0.f));
  CHECK_F44
    (a, f44(c,
     0.f, -2.f, -1.f, -3.f,
     -1.f, 2.f, 1.f, 3.f,
     -1.f, 0.f, 0.f, -2.f,
     -3.f, -2.f, -1.f, 0.f));

  f44_set(a, b);
  CHECK(f44_eq(a, b), 1);
  f44_add(a, a, f44_splat(c, FLT_EPSILON));
  CHECK(f44_eq_eps(a, b, FLT_EPSILON), 1);
  CHECK(f44_eq_eps(a, b, FLT_EPSILON * 0.9f), 0);

  f44_set
    (a, f44(c,
     2.f, 9.f, 8.f, 1.f,
     1.f, -2.f, 2.f, 1.f,
     1.f, -8.f, -4.f, 2.f,
     1.f, 3.f, 4.f, 2.f));
  CHECK(f44_det(a), 78.f);
  CHECK(f44_inverse(b, a), 78.f);
  f44_mulf44(dst, a, b);
  CHECK(f44_eq_eps
    (dst, f44(c,
     1.f, 0.f, 0.f, 0.f,
     0.f, 1.f, 0.f, 0.f,
     0.f, 0.f, 1.f, 0.f,
     0.f, 0.f, 0.f, 1.f),
     1.e-6f), 1);
  CHECK(f44_invtrans(a, a), 78.f);
  f44_transpose(a, a);
  CHECK_F44(a, b);

  return 0;
}
