/* Copyright (C) |Meso|Star> 2015-2016 (contact@meso-star.com)
 *
 * This software is a computer program whose purpose is to describe a
 * virtual 3D environment that can be ray-traced and sampled both robustly
 * and efficiently.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#if !defined(BUFFER_NAME) && !defined(BUFFER_DARRAY)

#ifndef S3D_BUFFER_H
#define S3D_BUFFER_H

#include <rsys/mem_allocator.h>
#include <rsys/ref_count.h>

#endif /* S3D_BUFFER_H */
#else
/*
 * Generate the buffer type with respect to the following macros:
 *  - BUFFER_NAME: name of the structure and prefix of the functions;
 *  - BUFFER_DARRAY: type of the dynamic array of the buffer;
 */
#if !defined(BUFFER_NAME) || !defined(BUFFER_DARRAY)
  #error "Missing macro definition"
#endif

#define BUFFER_FUNC__(Func) CONCAT(CONCAT(BUFFER_NAME, _), Func)
#define BUFFER_DARRAY_FUNC__(Func) CONCAT(CONCAT(BUFFER_DARRAY, _), Func)

struct BUFFER_NAME {
  BUFFER_DARRAY data;
  struct mem_allocator* allocator;
  ref_T ref;
};

/*******************************************************************************
 * Helper function
 ******************************************************************************/
static INLINE void
BUFFER_FUNC__(release__)(ref_T* ref)
{
  struct BUFFER_NAME* buffer;
  ASSERT(ref);
  buffer = CONTAINER_OF(ref, struct BUFFER_NAME, ref);
  BUFFER_DARRAY_FUNC__(release)(&buffer->data);
  MEM_RM(buffer->allocator, buffer);
}

/*******************************************************************************
 * Buffer function
 ******************************************************************************/
static INLINE res_T
BUFFER_FUNC__(create)
  (struct mem_allocator* allocator,
   struct BUFFER_NAME** out_buffer)
{
  struct BUFFER_NAME* buffer;
  ASSERT(allocator && out_buffer);

  buffer = (struct BUFFER_NAME*)MEM_CALLOC
    (allocator, 1, sizeof(struct BUFFER_NAME));
  if(!buffer) return RES_MEM_ERR;
  BUFFER_DARRAY_FUNC__(init)(allocator, &buffer->data);
  buffer->allocator = allocator;
  ref_init(&buffer->ref);
  *out_buffer = buffer;
  return RES_OK;
}

static INLINE void
BUFFER_FUNC__(ref_get)(struct BUFFER_NAME* buffer)
{
  ASSERT(buffer);
  ref_get(&buffer->ref);
}

static INLINE void
BUFFER_FUNC__(ref_put)(struct BUFFER_NAME* buffer)
{
  ASSERT(buffer);
  ref_put(&buffer->ref, BUFFER_FUNC__(release__));
}

#undef BUFFER_NAME
#undef BUFFER_DARRAY

#endif /* !BUFFER_NAME || !BUFFER_DARRAY */

