/* Copyright (C) |Meso|Star> 2015-2016 (contact@meso-star.com)
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "ssp.h"
#include "test_ssp_utils.h"

#include <rsys/float4.h>

#define NSAMPS 128

int
main(int argc, char** argv)
{
  struct ssp_rng* rng;
  struct mem_allocator allocator;
  float samps[NSAMPS][3];
  float A[3], B[3], C[3];
  float v0[3], v1[3], v2[3], v3[3], v4[3], v5[3];
  float plane[4];
  size_t counter[2];
  size_t nsteps;
  size_t i;
  (void)argc, (void)argv;

  mem_init_proxy_allocator(&allocator, &mem_default_allocator);

  CHECK(ssp_rng_create(&allocator, &ssp_rng_mt19937_64, &rng), RES_OK);

  FOR_EACH(i, 0, 3) {
    A[i] = (float)ssp_rng_uniform_double(rng, 0.0, 1.0);
    B[i] = (float)ssp_rng_uniform_double(rng, 0.0, 1.0);
    C[i] = (float)ssp_rng_uniform_double(rng, 0.0, 1.0);
  }

  f3_sub(v0, B, A);
  f3_sub(v1, C, A);
  f3_sub(v2, C, B);
  f3_minus(v3, v0);
  f3_minus(v4, v1);
  f3_minus(v5, v2);
  f3_cross(plane, v0, v1);
  plane[3] = -f3_dot(plane, C);

  FOR_EACH(i, 0, NSAMPS) {
    float tmp0[3], tmp1[3];
    float dot = 0.f;
    float area = 0.f;

    CHECK(ssp_ran_triangle_uniform(rng, A, B, C, samps[i]), samps[i]);
    CHECK(eq_epsf(f3_dot(plane, samps[i]), -plane[3], 1.e-6f), 1);

    f3_sub(tmp0, samps[i], A);
    dot = f3_dot(f3_cross(tmp0, tmp0, v0), f3_cross(tmp1, v1, v0));
    CHECK(signf(dot), 1.f );
    f3_sub(tmp0, samps[i], B);
    dot = f3_dot(f3_cross(tmp0, tmp0, v2), f3_cross(tmp1, v3, v2));
    CHECK(signf(dot), 1.f);
    f3_sub(tmp0, samps[i], C);
    dot = f3_dot(f3_cross(tmp0, tmp0, v4), f3_cross(tmp1, v5, v4));
    CHECK(signf(dot), 1.f);

    area = f3_len(tmp1) * 0.5f;
    CHECK(eq_epsf(1.f / area, ssp_ran_triangle_uniform_pdf(A, B, C), 1.e-6f), 1);
  }

  nsteps = 10000;
  counter[0] = counter[1] = 0;
  f3(A, -1.f, 0.f, 0.f);
  f3(B,  1.f, 0.f, 0.f);
  f3(C,  0.f, 1.f, 0.f);
  FOR_EACH(i, 0, nsteps) {
    ssp_ran_triangle_uniform(rng, A, B, C, samps[0]);
    if(samps[0][0] < 0.f)
      counter[0] += 1;
    else
      counter[1] += 1;
  }
  CHECK(labs((long)(counter[1] - counter[0])) < 100, 1);

  counter[0] = counter[1] = 0;
  FOR_EACH(i, 0, nsteps) {
    ssp_ran_triangle_uniform(rng, A, B, C, samps[0]);
    if(samps[0][1] < 1 - 1/sqrt(2))
      counter[0] += 1;
    else
      counter[1] += 1;
  }
  CHECK(labs((long)(counter[1] - counter[0])) < 100, 1);

  ssp_rng_ref_put(rng);
  check_memory_allocator(&allocator);
  mem_shutdown_proxy_allocator(&allocator);
  CHECK(mem_allocated_size(), 0);

  return 0;
}

